use std::fmt::Write;

use domain::base::rdata::ComposeRecordData;
use domain::base::wire::{Composer, ParseError};
use domain::base::{iana::Class, Name, ParseRecordData, ParsedName, RecordData, Record, Rtype, ToName, Ttl};
use domain::rdata;
use domain::dep::octseq::{Parser, Octets};

use crate::resources::dns::internal;
use crate::errors::Error;

#[derive(Debug)]
pub enum ProtoDnsError {
    RDataUnknown { input: String, field: String, rtype: String },
    NameParseError { input: String }
}

/* --------- A --------- */

impl From<domain::rdata::A> for internal::A {
    fn from(record_data: domain::rdata::A) -> Self {
        internal::A { address: record_data.addr() }
    }
}

impl TryFrom<internal::A> for domain::rdata::A {
    type Error = Error;

    fn try_from(record_data: internal::A) -> Result<Self, Self::Error> {
        Ok(domain::rdata::A::new(record_data.address))
    }
}

/* --------- AAAA --------- */

impl From<domain::rdata::Aaaa> for internal::Aaaa {
    fn from(record_data: domain::rdata::Aaaa) -> Self {
        internal::Aaaa { address: record_data.addr() }
    }
}

impl TryFrom<internal::Aaaa> for domain::rdata::Aaaa {
    type Error = Error;

    fn try_from(record_data: internal::Aaaa) -> Result<Self, Self::Error> {
        Ok(domain::rdata::Aaaa::new(record_data.address))
    }
}

/* --------- CNAME --------- */

impl<N: ToString> From<domain::rdata::Cname<N>> for internal::Cname {
    fn from(record_data: domain::rdata::Cname<N>) -> Self {
        internal::Cname { target: internal::Name::new(record_data.cname().to_string()) }
    }
}

impl TryFrom<internal::Cname> for domain::rdata::Cname<Name<Vec<u8>>> {
    type Error = Error;

    fn try_from(record_data: internal::Cname) -> Result<domain::rdata::Cname<Name<Vec<u8>>>, Self::Error> {
        let cname = record_data.target.to_string();

        let cname: Name<_> = cname.parse::<Name<_>>().map_err(|e| {
            Error::from(ProtoDnsError::RDataUnknown {
                input: cname,
                field: "target".into(),
                rtype: "CNAME".into(),
            }).with_cause(&e.to_string())
        })?;

        Ok(domain::rdata::Cname::new(cname))
    }
}

/* --------- MX --------- */

impl<N: ToString> From<domain::rdata::Mx<N>> for internal::Mx {
    fn from(record_data: domain::rdata::Mx<N>) -> Self {
        internal::Mx {
            preference: record_data.preference(),
            mail_exchanger: internal::Name::new(record_data.exchange().to_string()),
        }
    }
}

impl TryFrom<internal::Mx> for domain::rdata::Mx<Name<Vec<u8>>> {
    type Error = Error;

    fn try_from(record_data: internal::Mx) -> Result<domain::rdata::Mx<Name<Vec<u8>>>, Self::Error> {
        let mail_exchanger = record_data.mail_exchanger.to_string();

        let mail_exchanger: Name<_> = mail_exchanger.parse::<Name<_>>().map_err(|e| {
            Error::from(ProtoDnsError::RDataUnknown {
                input: mail_exchanger,
                field: "mail_exchanger".into(),
                rtype: "MX".into(),
            }).with_cause(&e.to_string())
        })?;

        Ok(domain::rdata::Mx::new(record_data.preference, mail_exchanger))
    }
}

/* --------- NS --------- */

impl<N: ToString> From<domain::rdata::Ns<N>> for internal::Ns {
    fn from(record_data: domain::rdata::Ns<N>) -> Self {
        internal::Ns {
            target: internal::Name::new(record_data.nsdname().to_string()),
        }
    }
}

impl TryFrom<internal::Ns> for domain::rdata::Ns<Name<Vec<u8>>> {
    type Error = Error;

    fn try_from(record_data: internal::Ns) -> Result<domain::rdata::Ns<Name<Vec<u8>>>, Self::Error> {
        let target = record_data.target.to_string();

        let target: Name<_> = target.parse::<Name<_>>().map_err(|e| {
            Error::from(ProtoDnsError::RDataUnknown {
                input: target,
                field: "target".into(),
                rtype: "NS".into(),
            }).with_cause(&e.to_string())
        })?;

        Ok(domain::rdata::Ns::new(target))
    }
}

/* --------- PTR --------- */

impl<N: ToString> From<domain::rdata::Ptr<N>> for internal::Ptr {
    fn from(record_data: domain::rdata::Ptr<N>) -> Self {
        internal::Ptr {
            target: internal::Name::new(record_data.ptrdname().to_string()),
        }
    }
}

impl TryFrom<internal::Ptr> for domain::rdata::Ptr<Name<Vec<u8>>> {
    type Error = Error;

    fn try_from(record_data: internal::Ptr) -> Result<domain::rdata::Ptr<Name<Vec<u8>>>, Self::Error> {
        let target = record_data.target.to_string();

        let target: Name<_> = target.parse::<Name<_>>().map_err(|e| {
            Error::from(ProtoDnsError::RDataUnknown {
                input: target,
                field: "target".into(),
                rtype: "PTR".into(),
            }).with_cause(&e.to_string())
        })?;

        Ok(domain::rdata::Ptr::new(target))
    }
}

/* --------- SOA --------- */

impl<N: ToString> From<domain::rdata::Soa<N>> for internal::Soa {
    fn from(record_rdata: domain::rdata::Soa<N>) -> Self {
        internal::Soa {
            primary_server: internal::Name::new(record_rdata.mname().to_string()),
            maintainer: internal::Name::new(record_rdata.rname().to_string()),
            refresh: record_rdata.refresh().as_secs(),
            retry: record_rdata.retry().as_secs(),
            expire: record_rdata.expire().as_secs(),
            minimum: record_rdata.minimum().as_secs(),
            serial: record_rdata.serial().into(),
        }
    }
}

impl TryFrom<internal::Soa> for domain::rdata::Soa<Name<Vec<u8>>> {
    type Error = Error;

    fn try_from(record_data: internal::Soa) -> Result<domain::rdata::Soa<Name<Vec<u8>>>, Self::Error> {
        let primary_server = record_data.primary_server.to_string();
        let primary_server: Name<_> = primary_server.parse::<Name<_>>().map_err(|e| {
            Error::from(ProtoDnsError::RDataUnknown {
                input: primary_server,
                field: "primary_server".into(),
                rtype: "SOA".into(),
            }).with_cause(&e.to_string())
        })?;

        let maintainer = record_data.maintainer.to_string();
        let maintainer: Name<_> = maintainer.parse::<Name<_>>().map_err(|e| {
            Error::from(ProtoDnsError::RDataUnknown {
                input: maintainer,
                field: "maintainer".into(),
                rtype: "SOA".into(),
            }).with_cause(&e.to_string())
        })?;

        Ok(domain::rdata::Soa::new(
            primary_server,
            maintainer,
            record_data.serial.into(),
            Ttl::from_secs(record_data.refresh),
            Ttl::from_secs(record_data.retry),
            Ttl::from_secs(record_data.expire),
            Ttl::from_secs(record_data.minimum),
        ))
    }
}

/* --------- SRV --------- */

impl<N: ToString> From<domain::rdata::Srv<N>> for internal::Srv {
    fn from(record_data: domain::rdata::Srv<N>) -> Self {
        internal::Srv {
            server: internal::Name::new(record_data.target().to_string()),
            priority: record_data.priority(),
            weight: record_data.weight(),
            port: record_data.port(),
        }
    }
}

impl TryFrom<internal::Srv> for domain::rdata::Srv<Name<Vec<u8>>> {
    type Error = Error;

    fn try_from(record_data: internal::Srv) -> Result<domain::rdata::Srv<Name<Vec<u8>>>, Self::Error> {
        let server = record_data.server.to_string();
        let server: Name<_> = server.parse::<Name<_>>().map_err(|e| {
            Error::from(ProtoDnsError::RDataUnknown {
                input: server,
                field: "server".into(),
                rtype: "SRV".into(),
            }).with_cause(&e.to_string())
        })?;


        Ok(domain::rdata::Srv::new(
            record_data.priority,
            record_data.weight,
            record_data.port,
            server
        ))
    }
}

/* --------- TXT --------- */

impl<O: AsRef<[u8]>> From<domain::rdata::Txt<O>> for internal::Txt {
    fn from(record_data: rdata::Txt<O>) -> Self {
        let concatenated_text: Vec<_> = record_data.iter()
            .flatten()
            .cloned()
            .collect();

        internal::Txt {
            text: concatenated_text
        }
    }
}

impl TryFrom<internal::Txt> for domain::rdata::Txt<Vec<u8>> {
    type Error = Error;

    fn try_from(record_data: internal::Txt) -> Result<domain::rdata::Txt<Vec<u8>>, Self::Error> {
        let txt: domain::rdata::Txt<_> = rdata::Txt::build_from_slice(&record_data.text).map_err(|e| {
            let text = record_data.text.iter()
                .fold(String::new(), |mut output, b| {
                    write!(output, "{:02X}", b).unwrap();
                    output
                });

            Error::from(ProtoDnsError::RDataUnknown {
                input: format!("0x{}", text),
                field: "text".into(),
                rtype: "TXT".into(),
            }).with_cause(&e.to_string())
        })?;


        Ok(txt)
    }
}

/* --------- ParsedRData  --------- */

pub enum ParsedRData<Name, Octs> {
    A(domain::rdata::A),
    Aaaa(domain::rdata::Aaaa),
    Cname(domain::rdata::Cname<Name>),
    Mx(domain::rdata::Mx<Name>),
    Ns(domain::rdata::Ns<Name>),
    Ptr(domain::rdata::Ptr<Name>),
    Soa(domain::rdata::Soa<Name>),
    Srv(domain::rdata::Srv<Name>),
    Txt(domain::rdata::Txt<Octs>),
}

impl<Name: ToString, Octs: AsRef<[u8]>> From<ParsedRData<Name, Octs>> for internal::RData {
    fn from(value: ParsedRData<Name, Octs>) -> Self {
        match value {
            ParsedRData::A(record_rdata) => internal::RData::A(record_rdata.into()),
            ParsedRData::Aaaa(record_rdata) => internal::RData::Aaaa(record_rdata.into()),
            ParsedRData::Cname(record_rdata) => internal::RData::Cname(record_rdata.into()),
            ParsedRData::Mx(record_rdata) => internal::RData::Mx(record_rdata.into()),
            ParsedRData::Ns(record_rdata) => internal::RData::Ns(record_rdata.into()),
            ParsedRData::Ptr(record_rdata) => internal::RData::Ptr(record_rdata.into()),
            ParsedRData::Soa(record_rdata) => internal::RData::Soa(record_rdata.into()),
            ParsedRData::Srv(record_rdata) => internal::RData::Srv(record_rdata.into()),
            ParsedRData::Txt(record_rdata) => internal::RData::Txt(record_rdata.into()),
        }
    }
}

impl TryFrom<internal::RData> for ParsedRData<Name<Vec<u8>>, Vec<u8>> {
    type Error = Error;

    fn try_from(value: internal::RData) -> Result<Self, Self::Error> {
        let rdata = match value {
            internal::RData::A(record_rdata) => ParsedRData::A(record_rdata.try_into()?),
            internal::RData::Aaaa(record_rdata) => ParsedRData::Aaaa(record_rdata.try_into()?),
            internal::RData::Cname(record_rdata) => ParsedRData::Cname(record_rdata.try_into()?),
            internal::RData::Mx(record_rdata) => ParsedRData::Mx(record_rdata.try_into()?),
            internal::RData::Ns(record_rdata) => ParsedRData::Ns(record_rdata.try_into()?),
            internal::RData::Ptr(record_rdata) => ParsedRData::Ptr(record_rdata.try_into()?),
            internal::RData::Soa(record_rdata) => ParsedRData::Soa(record_rdata.try_into()?),
            internal::RData::Srv(record_rdata) => ParsedRData::Srv(record_rdata.try_into()?),
            internal::RData::Txt(record_rdata) => ParsedRData::Txt(record_rdata.try_into()?),
        };
        Ok(rdata)
    }
}


impl<Name, Octs> ParsedRData<Name, Octs> {
    pub fn rtype(&self) -> Rtype {
        match self {
            ParsedRData::A(_) => Rtype::A,
            ParsedRData::Aaaa(_) => Rtype::AAAA,
            ParsedRData::Cname(_) => Rtype::CNAME,
            ParsedRData::Mx(_) => Rtype::MX,
            ParsedRData::Ns(_) => Rtype::NS,
            ParsedRData::Ptr(_) => Rtype::PTR,
            ParsedRData::Soa(_) => Rtype::SOA,
            ParsedRData::Srv(_) => Rtype::SRV,
            ParsedRData::Txt(_) => Rtype::TXT,
        }
    }
}

impl<Name, Octs> RecordData for ParsedRData<Name, Octs> {
    fn rtype(&self) -> Rtype {
        ParsedRData::rtype(self)
    }
}

impl<'a, Octs: Octets + ?Sized> ParseRecordData<'a, Octs> for ParsedRData<ParsedName<Octs::Range<'a>>, Octs::Range<'a>> {
    fn parse_rdata(
        rtype: Rtype,
        parser: &mut Parser<'a, Octs>,
    ) -> Result<Option<Self>, ParseError> {
        let record = match rtype {
            Rtype::A => ParsedRData::A(rdata::A::parse(parser)?),
            Rtype::AAAA => ParsedRData::Aaaa(rdata::Aaaa::parse(parser)?),
            Rtype::CNAME => ParsedRData::Cname(rdata::Cname::parse(parser)?),
            Rtype::MX => ParsedRData::Mx(rdata::Mx::parse(parser)?),
            Rtype::NS => ParsedRData::Ns(rdata::Ns::parse(parser)?),
            Rtype::PTR => ParsedRData::Ptr(rdata::Ptr::parse(parser)?),
            Rtype::SOA => ParsedRData::Soa(rdata::Soa::parse(parser)?),
            Rtype::SRV => ParsedRData::Srv(rdata::Srv::parse(parser)?),
            Rtype::TXT => ParsedRData::Txt(rdata::Txt::parse(parser)?),
            _ => return Ok(None)
        };

        Ok(Some(record))
    }
}

impl<Name: ToName, Octs: AsRef<[u8]>> ComposeRecordData for ParsedRData<Name, Octs> {
    fn rdlen(&self, compress: bool) -> Option<u16> {
        match self {
            ParsedRData::A(record_rdata) => record_rdata.rdlen(compress),
            ParsedRData::Aaaa(record_rdata) => record_rdata.rdlen(compress),
            ParsedRData::Cname(record_rdata) => record_rdata.rdlen(compress),
            ParsedRData::Mx(record_rdata) => record_rdata.rdlen(compress),
            ParsedRData::Ns(record_rdata) => record_rdata.rdlen(compress),
            ParsedRData::Ptr(record_rdata) => record_rdata.rdlen(compress),
            ParsedRData::Soa(record_rdata) => record_rdata.rdlen(compress),
            ParsedRData::Srv(record_rdata) => record_rdata.rdlen(compress),
            ParsedRData::Txt(record_rdata) => record_rdata.rdlen(compress),
        }
    }

    fn compose_rdata<Target: Composer + ?Sized>(
        &self,
        target: &mut Target,
    ) -> Result<(), Target::AppendError> {
        match self {
            ParsedRData::A(record_rdata) => record_rdata.compose_rdata(target),
            ParsedRData::Aaaa(record_rdata) => record_rdata.compose_rdata(target),
            ParsedRData::Cname(record_rdata) => record_rdata.compose_rdata(target),
            ParsedRData::Mx(record_rdata) => record_rdata.compose_rdata(target),
            ParsedRData::Ns(record_rdata) => record_rdata.compose_rdata(target),
            ParsedRData::Ptr(record_rdata) => record_rdata.compose_rdata(target),
            ParsedRData::Soa(record_rdata) => record_rdata.compose_rdata(target),
            ParsedRData::Srv(record_rdata) => record_rdata.compose_rdata(target),
            ParsedRData::Txt(record_rdata) => record_rdata.compose_rdata(target),
        }
    }

    fn compose_canonical_rdata<Target: Composer + ?Sized>(
        &self,
        target: &mut Target,
    ) -> Result<(), Target::AppendError> {
        match self {
            ParsedRData::A(record_rdata) => record_rdata.compose_canonical_rdata(target),
            ParsedRData::Aaaa(record_rdata) => record_rdata.compose_canonical_rdata(target),
            ParsedRData::Cname(record_rdata) => record_rdata.compose_canonical_rdata(target),
            ParsedRData::Mx(record_rdata) => record_rdata.compose_canonical_rdata(target),
            ParsedRData::Ns(record_rdata) => record_rdata.compose_canonical_rdata(target),
            ParsedRData::Ptr(record_rdata) => record_rdata.compose_canonical_rdata(target),
            ParsedRData::Soa(record_rdata) => record_rdata.compose_canonical_rdata(target),
            ParsedRData::Srv(record_rdata) => record_rdata.compose_canonical_rdata(target),
            ParsedRData::Txt(record_rdata) => record_rdata.compose_canonical_rdata(target),
        }
    }
}

/* --------- Records  --------- */

pub(crate) type RecordImpl = Record<
    Name<Vec<u8>>,
    ParsedRData<Name<Vec<u8>>,Vec<u8>>
>;


impl<Name: ToString, Oct: AsRef<[u8]>> From<Record<Name, ParsedRData<Name, Oct>>> for internal::Record {
    fn from(value: Record<Name, ParsedRData<Name, Oct>>) -> Self {
        internal::Record {
            name: internal::Name::new(value.owner().to_string()),
            ttl: value.ttl().as_secs(),
            rdata: internal::RData::from(value.into_data()),
        }
    }
}

impl TryFrom<internal::Record> for RecordImpl {
    type Error = Error;

    fn try_from(value: internal::Record) -> Result<Self, Self::Error> {
        let owner = value.name.to_string();
        let owner = owner.parse::<Name<_>>().map_err(|e|  {
            Error::from(ProtoDnsError::NameParseError {
                input: owner
            }).with_cause(&e.to_string())
        })?;

        let ttl = Ttl::from_secs(value.ttl);
        let data = value.rdata.try_into()?;
        Ok(Record::new(owner, Class::IN, ttl, data))
    }
}
